#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/select.h>
#include <sys/time.h>

#include <aisl/aisl.h>

#include "aisl.h"
#include "list.h"
#include "str-utils.h"
#include "handle.h"
#include "client.h"
#include "server.h"
#include "globals.h"
#include "stream.h"
#include "buffer.h"


struct aisl_instance
{
  list_t     * servers;
  #ifdef AISL_WITH_SSL
  list_t     * ssl;
  #endif
  list_t       servers;
  list_t       clients;
  list_t       callbacks;
  buffer_t     buffer;
  char       * last_error;
  int          iterator;
  int          stage;
  int          flags;
  uint32_t     accept_limit;
};


#ifdef AISL_WITH_SSL
static uint32_t m_instances = 0;
#endif

/* Initialization functions */

__attribute__ ((visibility ("default") ))
aisl_t
aisl_new( aisl_config_t * config )
{
  aisl_t self;

  #ifdef AISL_WITH_SSL
  if ((m_instances++) == 0)
  {
    SSL_load_error_strings();
    OpenSSL_add_ssl_algorithms();
  }
  #endif

  if ( !(self = calloc(1, sizeof(struct aisl_instance))) )
    goto finally;

  if ( !(self->servers = list_new(config->servers_spool_size)) )
    goto release;

  if ( !(self->clients = list_new(config->clients_spool_size)) )
    goto release;

  if ( !(self->callbacks = list_new(config->callbacks_spool_size)) )
    goto release;

  if ( !(self->buffer = buffer_new(config->initial_buffer_size)) )
    goto release;

  #ifdef AISL_WITH_SSL
  if ( !(self->ssl = list_new(config->ssl_spool_size)) )
    goto release;
  #endif

  self->accept_limit = config->clients_accept_limit;

  goto finally;

release:
  aisl_free(self);
  self = NULL;

finally:
  return self;
}


__attribute__ ((visibility ("default") ))
void
aisl_free( aisl_t self )
{
  if (self->clients)
    list_free(self->clients, (list_destructor_t) client_free );

  if (self->servers)
    list_free(self->servers, (list_destructor_t)  server_free );

  if (self->callbacks)
    list_free(self->callbacks, free);

  if (self->buffer)
    buffer_free(self->buffer);

  if (self->ssl)
    list_free(self->ssl, (list_destructor_t)  crypter_free );

  if (self->last_error)
    free(self->last_error);

  free(self);

  #ifdef AISL_WITH_SSL
  if ((--m_instances) == 0)
  {
    EVP_cleanup();
  }
  #endif
}


__attribute__ ((visibility ("default") ))
aisl_server_t
aisl_listen( aisl_instance_t    instance,
             const char       * address,
             uint16_t           port )
{
  aisl_server_t result;

  if ( (result = aisl_server_new(address, port)) != NULL )
  {
    result->instance = instance;

    if (list_append(instance->servers, result) == LIST_NAN)
    {
      server_free(result);
    }
  }

  return result;
}


static char *
get_response_begin(stream_t stream)
{
  char * r;
  client_t cli = CLIENT(ASTREAM(stream)->client);

  r = str_printf(
        "%s %d %s\r\n"
        "Server: AISL\r\n"
        "Connection: %s\r\n\r\n",
        aisl_http_version_to_string(cli->protocol),
        stream->response,
        aisl_http_response_to_string(stream->response),
        ((cli->flags & CLIENT_FLAG_KEEPALIVE) ? "keep-alive" : "close")
      );

  return r;
}

__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_response(aisl_stream_t stream, aisl_http_response_t  status_code,
                                    const char           *content_type,
                                    uint32_t              content_length)
{
  char  * pch;
  int     l;


  /* check if those headers were already sent */
  if (STREAM(stream)->state > STREAM_REQUEST_READY) return AISL_IDLE;

  STREAM(stream)->response = status_code;
  STREAM(stream)->c_type   = content_type;
  STREAM(stream)->c_length = content_length;

  if ( !(pch = get_response_begin(STREAM(stream))) )
    return AISL_MALLOC_ERROR;

  l = strlen(pch);
  STREAM(stream)->c_offset = l-2;

  buffer_clear(STREAM(stream)->buffer, content_length);

  l = buffer_add( STREAM(stream)->buffer, pch, l );
  free(pch);

  if (l == BUFFER_EOB) return AISL_MALLOC_ERROR;

  if (content_length)
  {
    if (!aisl_header_printf( stream, "Content-Length", "%u", content_length ))
      return AISL_MALLOC_ERROR;
  }

  if (content_type)
  {
    if (!aisl_header( stream, "Content-Type", content_type ))
      return AISL_MALLOC_ERROR;
  }

  STREAM(stream)->state = STREAM_RESPONSE_HEADER;

  return AISL_SUCCESS;
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_flush(aisl_stream_t stream)
{
  stream_t s = STREAM(stream); 
  if ( ! s->c_length )
  {
    s->c_length = s->buffer->size - s->c_offset - 2;
    if (!aisl_header_printf(stream, "Content-Length", "%u", s->c_length))
      return AISL_MALLOC_ERROR;
  }

  /*
  fprintf(stdout, "(%lu bytes)------->\n", STREAM(stream)->buffer->size);
  fwrite(STREAM(stream)->buffer->data, 1, STREAM(stream)->buffer->size, stdout);
  fprintf(stdout, "<------\n");
  */
  s->state = STREAM_RESPONSE_READY;

  s->flags |= STREAM_FLAG_OUTPUT_READY;

  return AISL_SUCCESS;
}


__attribute__ ((visibility ("default") ))
int
aisl_header(aisl_stream_t stream, const char *key, const char *value)
{
  int ret;
  char * pch;

  if ( (pch = str_printf("%s: %s\r\n", key, value)) != NULL )
  {
    ret = strlen(pch);
    if ( buffer_insert(
           STREAM(stream)->buffer,
           STREAM(stream)->c_offset,
           pch,
           ret
         ) == BUFFER_EOB )
    {
      ret = -1;
    }
    else
      STREAM(stream)->c_offset += ret;

    free(pch);
  }
  else
    ret = -1;

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_header_printf(aisl_stream_t stream, const char *key,
                                         const char *f_value, ...)
{
  int ret;

  va_list arg;
  va_start(arg, f_value);

  ret = aisl_header_vprintf(stream, key, f_value, arg);

  va_end(arg);

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_header_vprintf(aisl_stream_t stream, const char *key,
                                          const char *format,
                                          va_list     args)
{
  int ret;
  char * value;

  if ( (value = str_vprintf(format, args)) != NULL )
  {
    ret = aisl_header( stream, key, value );
    free(value);
  }
  else
    ret = -1;

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_printf(aisl_stream_t stream, const char *format, ...)
{
  va_list arg;
  va_start(arg, format);

  int result = aisl_vprintf(stream, format, arg);

  va_end(arg);

  /* No need to update length there, because vprintf do that
   * 
   * if (STREAM(stream)->c_length_unknown)
    STREAM(stream)->c_length += result;
    */


  return result;

}


__attribute__ ((visibility ("default") ))
int
aisl_vprintf(aisl_stream_t stream, const char *format, va_list args)
{
  int    result;
  char * r;

  if ( (r = str_vprintf(format, args)) != NULL)
  {
    result = strlen(r);
    if (buffer_add(STREAM(stream)->buffer, r, result) == BUFFER_EOB)
      result = -1;

    free(r);
  }
  else
    result = -1;

  return result;
}


__attribute__ ((visibility ("default") ))
int
aisl_write(aisl_stream_t stream, const char *data, int d_len)
{
  if (d_len < 0)
    d_len = strlen(data);

  if (buffer_add(STREAM(stream)->buffer, data, d_len) == BUFFER_EOB)
    d_len = -1;

  return d_len;
}


__attribute__ ((visibility ("default") ))
int
aisl_puts(const char *str, aisl_stream_t stream)
{
  return aisl_write( stream, str, strlen(str));
}


